@@ -1103,7 +1103,7 @@ class DebugUI(QObject):
         """
         self.ui.raise_()
         QApplication.processEvents()
-        if exceptionType is None:
+        if not exceptionType:
             E5MessageBox.critical(
                 self.ui, Program,
                 self.tr('An unhandled exception occured.'
@@ -1349,7 +1349,7 @@ class DebugUI(QObject):
         
         from .EditWatchpointDialog import EditWatchpointDialog
         dlg = EditWatchpointDialog(
-            (cond, temp, enabled, count, special), self)
+            (cond, temp, enabled, count, special), self.ui)
         if dlg.exec_() == QDialog.Accepted:
             cond, temp, enabled, count, special = dlg.getData()
             
