@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright (c) 2012 - 2015 Detlev Offenbach <detlev@die-offenbachs.de>
+# Copyright (c) 2012 - 2016 Detlev Offenbach <detlev@die-offenbachs.de>
 #
 
 """
@@ -246,10 +246,11 @@ class IrcWidget(QWidget, Ui_IrcWidget):
                             """ closed.</p>""")
                     .format(self.__server.getName()))
             if ok:
-                self.networkWidget.addServerMessage(
-                    self.tr("Info"),
-                    self.tr("Disconnecting from server {0}...").format(
-                        self.__server.getName()))
+                if self.__server is not None:
+                    self.networkWidget.addServerMessage(
+                        self.tr("Info"),
+                        self.tr("Disconnecting from server {0}...").format(
+                            self.__server.getName()))
                 self.__closeAllChannels()
                 self.__send("QUIT :" + self.__quitMessage)
                 self.__socket and self.__socket.flush()
